#ifndef CUDAMATH_H
#define CUDAMATH_H

#include <cutil_inline.h>



template <class T> T peekValue(T* tab,unsigned int i) {
	T result;
	cutilSafeCall(cudaMemcpy(&result,tab+i,sizeof(T),cudaMemcpyDeviceToHost));
	return result;
}

template <unsigned int blockSize> __global__ void reduceKernel(float *g_idata, float *g_odata, unsigned int n) {
	extern __shared__ float sdata[];

	unsigned int tid = threadIdx.x;
	unsigned int i = blockIdx.x*(blockSize*2) + tid;
	unsigned int gridSize = blockSize*2*gridDim.x;
	sdata[tid] = 0.0f;
		
	  while (i+blockSize < n) { sdata[tid] += g_idata[i] + g_idata[i+blockSize];  i += gridSize;  } 
	  if (i < n ) { sdata[tid] += g_idata[i]; } 
	__syncthreads();

	if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
	if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
	if (blockSize >= 128) { if (tid < 64)  { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
    
	if (tid < 32) {
		if (blockSize >=  64) sdata[tid] += sdata[tid + 32];
		if (blockSize >=  32) sdata[tid] += sdata[tid + 16];
		if (blockSize >=  16) sdata[tid] += sdata[tid +  8];
		if (blockSize >=    8) sdata[tid] += sdata[tid +  4];
		if (blockSize >=    4) sdata[tid] += sdata[tid +  2];
		if (blockSize >=    2) sdata[tid] += sdata[tid +  1];
	}

	if (tid == 0) g_odata[blockIdx.x] = sdata[0];

}


template <unsigned int blockSize> __host__ float lossyReduce(float *d_idata, unsigned int n) {
	
	//Allocate buffer
	uint tempBufferSize=min(65535,intDivCeil(n,blockSize*2)); //WARNING 65535 czy 655536? sprawdzić
	
	
	float *d_tempBuffer;
	
	cutilSafeCall(cudaMalloc((void**)&d_tempBuffer,sizeof(float)*tempBufferSize));
	
	//First step - reduce the input data to the tempBuffer
	
	
	//printf("n %d tempBufferSize %d blockSize %d sizeof %d\n",n,tempBufferSize,blockSize,sizeof(float));
	
	reduceKernel<blockSize><<<tempBufferSize,blockSize,sizeof(float)*blockSize>>>(d_idata,d_tempBuffer,n);
	//CUT_CHECK_ERROR("Kernel execution failed");
	
	/*for (int i=0;i<tempBufferSize;i++) {
		printf("%f \n",peekValue(d_idata,i));
	}*/
	
	
	//Second step - use ping-pong technique to reduce the tempBuffer to a single value
	
	float *d_out=d_tempBuffer;
	float *d_in=d_idata;
	float *temp;
	uint oldTempBufferSize;
	
	while (tempBufferSize>1) {
		temp=d_out;
		d_out=d_in;
		d_in=temp;
		
		
		oldTempBufferSize=tempBufferSize;
		tempBufferSize=intDivCeil(tempBufferSize,2*blockSize); 
		
		reduceKernel<blockSize><<<tempBufferSize,blockSize,sizeof(float)*blockSize>>>(d_in,d_out,oldTempBufferSize);
	}
	
	//Last step - copy the result to the output variable and free the tempBuffer;
	
	float result=peekValue(d_out,0);
	
	cutilSafeCall(cudaFree(d_tempBuffer));
	
	return result;
}



#endif