#include "lscvmulticudakernels.h"
#include "cutil_inline.h"
#include "cudamath.h"

__device__ __constant__ float dcG[256];


__device__ __inline__ float th(float ckkh, float ckh, float expcomp) {
	return ckkh*sqrt(expcomp)-ckh*expcomp;
}

__device__ inline float ceilSquareRoot(uint x) {
	float a=sqrtf(x);
	//return round(a)*round(a)==x?round(a):ceil(a);
	return round(a)*round(a)==x?round(a):a;
}



template <int side,int d> __global__ void calcKHComponents(float *data, float* out, int n, float ckkh, float ckh) {
	extern __shared__ float Mi[];

	//float* Mi=base;
	float* Mj=Mi+side*side;
	__shared__ float sumBuf[side*side];
	__shared__ float ldcG[side*side];
	__shared__ float partt[d];

	int bx=blockIdx.x+blockIdx.y*gridDim.x;

	int iBegin=ceil((ceilSquareRoot((bx<<3)+9)-3.0f)/2.0f);
	int jBegin=(bx-((iBegin*(iBegin+1))>>1))*side;
	iBegin*=side;

	int tx=threadIdx.x%side;
	int ty=threadIdx.x/side;

	//int iBegin=i*side;
	//int jBegin=j*side;

	float part=0.0f;
	float res=0.0f;
	
	Mj[threadIdx.x]=data[min(jBegin+tx,n-1)+min(ty,d-1)*n]; 
	Mi[threadIdx.x]=data[min(iBegin+tx,n-1)+min(ty,d-1)*n];
	sumBuf[threadIdx.x]=0.0f;
	ldcG[threadIdx.x]=dcG[threadIdx.x];

	__syncthreads();
	
	if ((jBegin+tx)<(iBegin+ty)) {
		//if ((jBegin+tx)<n && (iBegin+tx)<n ) {
			
			for (int p=0;p<d;p++) {
				part=0.0f;
				
				for (int k=0;k<d;k++) {
					part+=ldcG[k*d+p]*(Mi[k*side+ty]-Mj[k*side+tx]);
				}
				
				res+=part*(Mi[p*side+ty]-Mj[p*side+tx]);
			}
			sumBuf[threadIdx.x]=th(ckkh,ckh,1/sqrt(exp(res)));
		//} 	
		
	}
	
	__syncthreads();

	if (side*side >= 512) { if (threadIdx.x < 256) { sumBuf[threadIdx.x] += sumBuf[threadIdx.x + 256]; } __syncthreads(); }
	if (side*side >= 256) { if (threadIdx.x < 128) { sumBuf[threadIdx.x] += sumBuf[threadIdx.x + 128]; } __syncthreads(); }
	if (side*side >= 128) { if (threadIdx.x < 64)  { sumBuf[threadIdx.x] += sumBuf[threadIdx.x +  64]; } __syncthreads(); }
    
	if (threadIdx.x < 32) {
		volatile float* smem = sumBuf;
		if (side*side >=  64) smem[threadIdx.x] += smem[threadIdx.x + 32];
		if (side*side >=  32) smem[threadIdx.x] += smem[threadIdx.x + 16];
		if (side*side >=  16) smem[threadIdx.x] += smem[threadIdx.x +  8];
		if (side*side >=    8) smem[threadIdx.x] += smem[threadIdx.x +  4];
		if (side*side >=    4) smem[threadIdx.x] += smem[threadIdx.x +  2];
		if (side*side >=    2) smem[threadIdx.x] += smem[threadIdx.x +  1];
	}

	if (threadIdx.x == 0) out[bx] = sumBuf[0];

	
}


float *gpuAllocTempArrayForGH(int n,int side) {
	float *o;
	cutilSafeCall(cudaMalloc((void**)&o,sizeof(float)*intDivCeil(n,16)*intDivCeil(n+1,16)/2));
	return o;
}

__host__ float* gpuAllocReduceBuffer(int blockSize, int n) {
	float *d_tempBuffer;
	//Allocate buffer
	uint tempBufferSize=min(65535,intDivCeil(n,blockSize*2)); //WARNING 65535 czy 655536? sprawdzić
	cutilSafeCall(cudaMalloc((void**)&d_tempBuffer,sizeof(float)*tempBufferSize));
	return d_tempBuffer;
}


//float gH_gpu_multi(float *X, float *H, int d, int n, float *ghtemparray, float *khtemparray, float *sumTempArray) {


float calcGH(HAlgorithm<float>* a,float *data, float *H, int n, int d) {

	TIMESTAMP totalStart,totalEnd;
	TIMESTAMP transferStart,transferEnd;
	TIMESTAMP khStart,khEnd;
	TIMESTAMP reduceStart,reduceEnd;
	//int blockSize=512;
	//unsigned int timer;
	//cutilCheckError(cutCreateTimer(&timer));
	//cutilCheckError(cutStartTimer(timer));
	
	totalStart=getTimer();
	float* ghtemparray=gpuAllocTempArrayForGH(n,16);
	float* sumTempArray=gpuAllocReduceBuffer(n,512);
	float* X;
	
	cutilSafeCall(cudaMalloc((void**)&X,sizeof(float)*n*d));
	
	transferStart=getTimer();
	cutilSafeCall(cudaMemcpy(X,data,sizeof(float)*n*d,cudaMemcpyHostToDevice));
	transferEnd=getTimer();
	
	float *G;
	float det;
	
	//printf("n %d , d %d \n",n,d);
	
	//printf("Inverting matrix\n");
	invertMatrixAndFindDeterminant(d, H, &G, &det);
	//printf("Copying matrix to constant memory\n");
	cutilSafeCall(cudaMemcpyToSymbol(dcG, G, sizeof(float)*d*d,0,cudaMemcpyHostToDevice));
	
	
	float invsqrtdetH=pow(det,-0.5);
	float ckkh= pow(4*M_PI,-0.5*d)*invsqrtdetH;
	float ckh= 2*pow(2*M_PI,-0.5*d)*invsqrtdetH;
	
	/*
	int blocks=intDivCeil(n,16)*(intDivCeil(n,16)+1)/2;
	int blocksY=intDivCeil(blocks,65535);
	

	blocks=intDivCeil(blocks,blocksY);
	
	
	printf("old blocks %d\n",blocks);
	printf("old blocksY %d\n",blocksY);
	*/
	
	int v1=intDivCeil(n,16);
	int v2=v1+1;
	
	if ((v1&1)==0) v1>>=1; else v2>>=1;
	
	int blocks=v1;
	int blocksY=v2;
	
	//printf("blocks %d\n",blocks);
	//printf("blocksY %d\n",blocksY);
	//printf("blocks %d\n",blocks);
	//printf("Finding KH components \n");
	khStart=getTimer();
	dim3 grid=dim3(blocks,blocksY);
	switch (d) {
		case 1:  calcKHComponents<16,1><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 2:  calcKHComponents<16,2><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 3:  calcKHComponents<16,3><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 4:  calcKHComponents<16,4><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 5:  calcKHComponents<16,5><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 6:  calcKHComponents<16,6><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 7:  calcKHComponents<16,7><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 8:  calcKHComponents<16,8><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 9:  calcKHComponents<16,9><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 10: calcKHComponents<16,10><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 11: calcKHComponents<16,11><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 12: calcKHComponents<16,12><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 13: calcKHComponents<16,13><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 14: calcKHComponents<16,14><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 15: calcKHComponents<16,15><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
		case 16: calcKHComponents<16,16><<<grid,16*16,16*16*2*sizeof(float)>>>(X, ghtemparray,n,ckkh,ckh);break;
	}
	cutilSafeCall(cudaThreadSynchronize());
	khEnd=getTimer();
	
	CUT_CHECK_ERROR("Exponents kernel execution failed");
	
	//float *rob=new float[n*(n-1)/2];
	//cudaMemcpy(rob,ghtemparray,(n*(n-1)/2)*sizeof(float),cudaMemcpyDeviceToHost);
	
	/*for (int i=0;i<blocks*blocksY;i++) {
		printf(">>> %f \n",rob[i]);
	}*/
	
	/*for (int i=0;i<256;i++) {
		printf("cuda %f \n",rob[i]);
	}*/
	
	//printf("det: %f \n",det);
	
	reduceStart=getTimer();
	float sum=lossyReduce<512>(ghtemparray,intDivCeil(n,16)*(intDivCeil(n,16)+1)/2);          //reduceKhComponents<512>(ghtemparray, sumTempArray, n*(n-1)/2,ckkh,ckh);
	reduceEnd=getTimer();
	//printf("suma %f \n",sum);
	
	delete []G;
	
	cutilSafeCall(cudaFree(ghtemparray));
	cutilSafeCall(cudaFree(sumTempArray));
	cutilSafeCall(cudaFree(X));

	//cutilCheckError(cutStopTimer(timer));
	//double time=cutGetTimerValue(timer);
	//printf("\t partial %f \n",time);
	//cutilCheckError(cutDeleteTimer(timer));
	
	float invrk=pow(2.0,d)*pow(M_PI,0.5*d)*sqrt(det);
	float result=2*sum/(n*n)+1/(n*invrk);
	
	totalEnd=getTimer();
	
	a->addMeasure("LSCVMULTI_CUDA_TRAN",transferEnd-transferStart);
	a->addMeasure("LSCVMULTI_CUDA_KH",khEnd-khStart);
	a->addMeasure("LSCVMULTI_CUDA_REDUCE",reduceEnd-reduceStart);
	a->addMeasure("LSCVMULTI_CUDA_TOTAL",totalEnd-totalStart);
	
	return result;
}
