#include "lscvmultisimple.h"



float *LSCVMultiSimple::allocTempArrayForGH(int n) {
  return new float[n*(n-1)/2];
}

float *LSCVMultiSimple::allocTempArrayForKH(int d) {
  return new float[d];
}

/**
 * \brief Zwraca odwrotność wartości funkcji R(K) (wzór 28)
 * \param d liczba wierszy w macierzy X (wymiarowość)
 * \param sqrtdetH - pierwiastek kwadratowy z wyznacznika H
 * \return odwrotność wartości funkcji R(K)
 */
float LSCVMultiSimple::invrk(int d,float sqrtdetH) {
  return pow(2.0,d)*pow(M_PI,0.5*d)*sqrtdetH;
}


float  LSCVMultiSimple::KH(float *Y, float* G,int d, int n,int i, int j,float* temp) {
  /*for (int k=0;k<d;k++) {
    temp[k]=0;
    for (int l=0;l<d;l++) {
      temp[k]+=(X[i+n*l]-X[j+n*l])*G[k+d*l];
    }
  }
  
  float w2=0;
  
  for (int l=0;l<d;l++) {
    w2+=(X[i+n*l]-X[j+n*l])*temp[l];
  }*/
  
  //printf("w2simple %f \n",w2);
  
  for (int k=0;k<d;k++) {
    temp[k]=0;
    for (int l=0;l<d;l++) {
      temp[k]+=(Y[l+d*i]-Y[j*d+l])*G[k+d*l];
    }
  }
  
  float w2=0;
  
  for (int l=0;l<d;l++) {
    w2+=(Y[i*d+l]-Y[j*d+l])*temp[l];
  }
  
  //printf("w2simple %f \n",w2);
  return w2;
}



float LSCVMultiSimple::findGH(float *H) {
	int n=getN();
	int d=getD();
	float *G;
	float *X=getData();
	float det;
	float kh;
	float kkh;
	float* ghtemparray=allocTempArrayForGH(n);
	float* khtemparray=allocTempArrayForKH(d);
  
	TIMESTAMP start=getTimer();
	unsigned int total=n*(n-1)/2;
	invertMatrixAndFindDeterminant(d, H, &G, &det);
	
	float *Y=transposeMatrix(X,d,n);
	
	float sdet=sqrt(det);
	
	float c1=pow(4.0*M_PI,-(float)d/2.0)/sdet;
	float c2=2*pow(2.0*M_PI,-(float)d/2.0)/sdet;
  
	for (int i=0;i<n;i++) {
		for (int j=i+1;j<n;j++) {
			float w=KH(Y,G,d,n,i,j,khtemparray);
			
			kkh=c1*exp(-0.25*w);
			kh=c2*exp(-0.5*w);
			
			//printf("khsim %f \n",kkh-2*kh);
			ghtemparray[j-i-1+(2*n-i-1)*i/2]=kkh-kh;
		}
	}
  
	
	float res=stableUnoptimizedSum<false>(ghtemparray,total);
 
	//printf("res sim %f \n",res);
	res=2*res/(float)(n*n)+1/((float)n*invrk(d,sdet));
	TIMESTAMP end=getTimer();
	
	addMeasure("LSCVMULTI_SIMPLE_TOTAL",end-start);
	
	delete []G;
	delete []ghtemparray;
	delete []khtemparray;
	delete []Y;
  
	return res;
}

float LSCVMultiSimple::findGHForEst() {
	int d=getD();
	int n=getN();
	float * data=getData();
	float *cov=covarianceMatrix(data,d,n);
	float coeff=pow(4.0/(float)(d+2),1.0/(float)(d+4))*pow(n,-1.0/(float)(d+4));
	coeff=coeff*coeff;
	multMatrix(cov,d,d,coeff);
	float result=findGH(cov);
	delete []cov;
	return result;

}

void LSCVMultiSimple::startExp() {
	findGHForEst();
}