#include "lscvmultisse.h"


/**
 * \brief Zwraca odwrotność wartości funkcji R(K) (wzór 28)
 * \param d liczba wierszy w macierzy X (wymiarowość)
 * \param sqrtdetH - pierwiastek kwadratowy z wyznacznika H
 * \return odwrotność wartości funkcji R(K)
 */
float LSCVMultiSSE::invrk(int d,float sqrtdetH) {
  return pow(2.0,d)*pow(M_PI,0.5*d)*sqrtdetH;
}

int LSCVMultiSSE::setThreads(int threads) {
	this->threads=threads;
}
__m128 LSCVMultiSSE::k(__m128 v,void *d) {
	__m128 sqrtdeth=((__m128*)d)[0];
	__m128 twotwopi=((__m128*)d)[1];
	__m128 fourpi=((__m128*)d)[2];
	
	__m128 e1=exp_ps(_mm_div_ps(v,_mm_set1_ps(-4.0f)));
	__m128 e2=exp_ps(_mm_div_ps(v,_mm_set1_ps(-2.0f)));
	
	return _mm_div_ps(_mm_sub_ps(_mm_mul_ps(fourpi,e1),_mm_mul_ps(twotwopi,e2)),sqrtdeth);
}

float LSCVMultiSSE::findGH(float *H) {
	int n=getN();
	int d=getD();
	
	float *X=getData();
	float det;
	float *G;
	
  
	TIMESTAMP start=getTimer();
	unsigned int total=n*(n-1)/2;
	invertMatrixAndFindDeterminant(d, H, &G, &det);
  
	float sqrtdet=sqrt(det);
	float twotwopi=2*pow(2*M_PI,-(float)d/2.0f);
	float fourpi=pow(4*M_PI,-(float)d/2.0f);
	
	__m128 data[3];
	
	data[0]=_mm_set1_ps(sqrtdet);
	data[1]=_mm_set1_ps(twotwopi);
	data[2]=_mm_set1_ps(fourpi);
	
	
  
	float res= vecvecMatBigSSEParallelFunSum<LSCVMultiSSE::k>(X,G,n,d,threads,(void*) data);
	//printf("res sse %f \n",res);
	
	res=2*res/(float)(n*n)+1/((float)n*invrk(d,sqrtdet));
	TIMESTAMP end=getTimer();
	
	addMeasure("LSCVMULTI_SSE_TOTAL",end-start);
	
	delete []G;
  
	return res;
}

float LSCVMultiSSE::findGHForEst() {
	int d=getD();
	int n=getN();
	float * data=getData();
	float *cov=covarianceMatrix(data,d,n);
	float coeff=pow(4.0/(float)(d+2),1.0/(float)(d+4))*pow(n,-1.0/(float)(d+4));
	coeff=coeff*coeff;
	multMatrix(cov,d,d,coeff);
	float result=findGH(cov);
	delete []cov;
	return result;

}

void LSCVMultiSSE::startExp() {
	findGHForEst();
}