#include "lscvsimple.h"


void LSCVSimple::setStep(float step) {
	this->step=step;
}

float LSCVSimple::getStep() {
	return step;
}

float  LSCVSimple::kstar2(float &v,int i,void *p) {
	float k;
	float k2;

	float h2=((float*)p)[0];
	float C1=((float*)p)[1];
	float C2=((float*)p)[2];
	
	k=exp(-v/(2.0f*h2))/C1;
	k2=exp(-v/(4.0f*h2))/C2;

	return k2-2*k;
}


float LSCVSimple::gh(float h, unsigned int n, unsigned int d, float C1, float C2,  float* exponents, float *temp) {
	float sum=0;

	unsigned int total=n*(n-1)/2;

	//float h2=h*h;
	
	float data[]={h*h,C1,C2};
	
	//template <float (*F)(float &,int i,void *),bool preserve> float stableUnoptimizedSumFun(float *t,int n,void *data)
	sum=stableUnoptimizedSumFun<LSCVSimple::kstar2,true>(exponents,total,data);
	//printf("sum %f \n",sum);
	sum=(2*sum/((float)(n*n))+1/(n*C2))/pow(h,d);
	/*for (int i=0;i<total;i++) {
			temp[i]=kstar(h2,n,C1,C2,exponents,i);
	}

	unsigned int n2=roundp2(total); //zaokrąglij do pierwszej większej potęgi 2.

	for (int step=n2/2;step>=1;step=step/2) {
		for (int i=0;i<step;i++) {
			if ((i+step)<total) temp[i]=temp[i]+temp[i+step];
		}
	}

	
	printf("sum %f \n",temp[0]);
 	sum=(2*temp[0]/((float)(n*n))+1/(n*C2))/pow(h,d);*/

	return sum;
}


float LSCVSimple::findH() {
	unsigned int n=getN();
	unsigned int d=getD();
	float step=getStep();
	unsigned int timerTotal=0;
	unsigned int timerTemp=0;
	float *data=getData();

	TIMESTAMP start=getTimer();
	
	//Oblicz macierz kowariancji
	float *cov=covarianceMatrix(data,d,n);

	float detCov;
	float *invCov;

	//Odwróć macierz kowariancji i oblicz jej wyznacznik
	invertMatrixAndFindDeterminant(d,cov,&invCov,&detCov);

	//Oblicz wyrażenia pomocnicze
	float C1=pow(2*M_PI,(float)d/2.0f)*sqrt(detCov);
	float C2=pow(4*M_PI,(float)d/2.0f)*sqrt(detCov);

	//Oblicz wyznaczniki
	float* exponents;
	
	TIMESTAMP exponentsStart=getTimer();
	exponents=findExponents(d,n,data,invCov);
	TIMESTAMP exponentsEnd=getTimer();
	addMeasure("LSCV_SIMPLE_EXPONENTS",exponentsEnd-exponentsStart);
	
	//Oblicz wyrażenia pomocnicze
	float rk_mik2_inv=d*d*pow(2,d)*pow(M_PI,(float)d/2.0f);
	float rf_inv=(pow(2,d+2)*pow(M_PI,(float)d/2.0f))/((float)(d*(d+2)));
	float h0=pow((float)d*rf_inv/(rk_mik2_inv*(float)n),1/(d+4.0f));


	float h0_min=h0/4.0f;
	float h0_max=h0*4.0f;

	//Liczba wartości g(h) do policzenia
	int steps=150;//ceil((h0_max-h0_min)/step); //WARNING rozwiązanie tymczasowe, ignoruje inne procedury
	step=(h0_max-h0_min)/steps;

	float *temp=new float[n*(n-1)/2];

	float ghest_min=gh(h0_min,n,d,C1,C2,exponents,temp);
	float hmin=h0_min;

	for (int i=1;i<steps;i++) {
		float ghest=gh(h0_min+step*i,n,d,C1,C2,exponents,temp);
		if (ghest<ghest_min) {
			hmin=h0_min+step*i;
			ghest_min=ghest;
		}
	}
	

	delete []temp;
	delete []invCov;
	delete []cov;
	delete []exponents;

	TIMESTAMP end=getTimer();
	
	
	addMeasure("LSCV_SIMPLE_STEPS",steps);
	addMeasure("LSCV_SIMPLE_TOTAL",end-start);

	return hmin;
}

void LSCVSimple::startExp() {
	findH();
}
