#include "lscvsse.h"


void LSCVSSE::setStep(float step) {
	this->step=step;
}

float LSCVSSE::getStep() {
	return step;
}

float  LSCVSSE::kstar(float &v,int i,void *p) {
	float k;
	float k2;

	float h2=((float*)p)[0];
	float C1=((float*)p)[1];
	float C2=((float*)p)[2];
	
	k=exp(-v/(2.0f*h2))/C1;
	k2=exp(-v/(4.0f*h2))/C2;

	return k2-2*k;
}

__m128 LSCVSSE::kstarsse(__m128 &x,int i,void *v) {
	__m128 k;
	__m128 k2;

	float h2=((float*)v)[0];
	float C1=((float*)v)[1];
	float C2=((float*)v)[2];
	
	__m128 h2m=_mm_set1_ps(h2);
	
	
	k=_mm_div_ps(exp_ps(_mm_div_ps(x,_mm_mul_ps(h2m,_mm_set1_ps(-2.0f)))),_mm_set1_ps(C1));
	k2=_mm_div_ps(exp_ps(_mm_div_ps(x,_mm_mul_ps(h2m,_mm_set1_ps(-4.0f)))),_mm_set1_ps(C2));
	

	return _mm_sub_ps(k2,_mm_add_ps(k,k));
}

void LSCVSSE::setThreads(int threads) {
	this->threads=threads;
}

void LSCVSSE::setCachesize(int cachesize) {
	this->cachesize=cachesize;
}


float LSCVSSE::gh(float h, unsigned int n, unsigned int d, float C1, float C2,  float* exponents, float *temp) {
	unsigned int total=n*(n-1)/2;
	float data[]={h*h,C1,C2};
	float sum=multiThreadedAggregateOMPFun<LSCVSSE::kstarsse,LSCVSSE::kstar,true>(exponents,total,data,threads,cachesize);
	sum=(2*sum/((float)(n*n))+1/(n*C2))/pow(h,d);

	return sum;
}


float LSCVSSE::findH() {
	unsigned int n=getN();
	unsigned int d=getD();
	float step=getStep();
	unsigned int timerTotal=0;
	unsigned int timerTemp=0;
	float *data=getData();

	TIMESTAMP start=getTimer();
	
	//Oblicz macierz kowariancji
	float *cov=covarianceMatrix(data,d,n);

	float detCov;
	float *invCov;

	//Odwróć macierz kowariancji i oblicz jej wyznacznik
	invertMatrixAndFindDeterminant(d,cov,&invCov,&detCov);

	//Oblicz wyrażenia pomocnicze
	float C1=pow(2*M_PI,(float)d/2.0f)*sqrt(detCov);
	float C2=pow(4*M_PI,(float)d/2.0f)*sqrt(detCov);

	//Oblicz wyznaczniki
	float* exponents;
	//exponents=findExponents(d,n,data,invCov);
	exponents=new float[n*(n-1)/2];
	vecvecMatBigSSEParallel(data,invCov,n,d,exponents,threads);
	//exponents=
	
	//Oblicz wyrażenia pomocnicze
	float rk_mik2_inv=d*d*pow(2,d)*pow(M_PI,(float)d/2.0f);
	float rf_inv=(pow(2,d+2)*pow(M_PI,(float)d/2.0f))/((float)(d*(d+2)));
	float h0=pow((float)d*rf_inv/(rk_mik2_inv*(float)n),1/(d+4.0f));


	float h0_min=h0/4.0f;
	float h0_max=h0*4.0f;

	//Liczba wartości g(h) do policzenia
	//int steps=ceil((h0_max-h0_min)/step);
	int steps=150; //WARNING rozwiązanie tymczasowe, ignoruje inne procedury
	step=(h0_max-h0_min)/steps;


	float *temp=new float[n*(n-1)/2];

	float ghest_min=gh(h0_min,n,d,C1,C2,exponents,temp);
	float hmin=h0_min;

	for (int i=1;i<steps;i++) {
		float ghest=gh(h0_min+step*i,n,d,C1,C2,exponents,temp);
		if (ghest<ghest_min) {
			hmin=h0_min+step*i;
			ghest_min=ghest;
		}
	}
	

	delete []temp;
	delete []invCov;
	delete []cov;
	delete []exponents;

	TIMESTAMP end=getTimer();
	
	addMeasure("LSCV_SSE_TOTAL",end-start);
	addMeasure("LSCV_SSE_STEPS",steps);

	return hmin;
}

void LSCVSSE::startExp() {
	findH();
}
