#include "pluginsimple.h"
#include "pluginsse.h"
#include "plugincuda.h"
#include "lscvsimple.h"
#include "lscvsse.h"
#include "lscvcuda.h"
#include "lscvmultisimple.h"
#include "lscvmultisse.h"
#include "lscvmulticuda.h"
#include "cudautils.h"
#include "halgorithm.h"
#include <stdio.h>
#include <stdlib.h>

__m128 f(__m128 a) {
	return _mm_mul_ps(a,a);
}

void experiment(int n, int d, HAlgorithm<float> **algs,int nalg,FILE *plik,int repetitions) {
	float *X=new float[n*d];
	
	//for (int i=0;i<n*d;i++) X[i]=2*((float)rand()/(float)RAND_MAX)-1;
	
	for (int i=0;i<nalg;i++) {
		printf("%d %d\t",n,d);
		for (int j=0;j<repetitions;j++) {
			for (int k=0;k<n*d;k++) X[k]=2*((float)rand()/(float)RAND_MAX)-1;
			algs[i]->setData(X,n,d);
			algs[i]->startExp();
			for (int j=0;j<algs[i]->getMeasureCount();j++) {
				fprintf(plik,"%d %d %s\t %lld\n",n,d,algs[i]->getMeasureName(j).c_str(),algs[i]->getMeasureTime(j));
				printf("%s\t %lld\t",algs[i]->getMeasureName(j).c_str(),algs[i]->getMeasureTime(j));
			}
			//fprintf(plik,"\n");
			printf("\n");
			fflush(plik);
			algs[i]->clearStats();
		}
	}
	
	delete []X;
}

int main(int argc, char** argv) {
	
	
	eatDelay();
	
	
	HAlgorithm<float>* pluginAlgs[3];
	
	PluginSimple ps;
	PluginSSE pss;
	pss.setThreads(8);
	pss.setCachesize(8192);
	PluginCUDA psc;
	
	pluginAlgs[0]=&ps;
	pluginAlgs[1]=&psc;
	pluginAlgs[2]=&pss;
	
	HAlgorithm<float>* lscvAlgs[3];
	
	LSCVSimple ls;
	ls.setStep(0.01);
	LSCVSSE lss;
	lss.setStep(0.01);
	lss.setThreads(8);
	lss.setCachesize(8192);
	LSCVCUDA lsc;
	lsc.setStep(0.01);
	
	lscvAlgs[0]=&ls;
	lscvAlgs[1]=&lss;
	lscvAlgs[2]=&lsc;
	
	HAlgorithm<float>* lscvMultiAlgs[3];
	
	LSCVMultiSimple lms;
	LSCVMultiSSE lme;
	lme.setThreads(8);
	LSCVMultiCUDA lmc;
	
	lscvMultiAlgs[0]=&lms;
	lscvMultiAlgs[1]=&lme;
	lscvMultiAlgs[2]=&lmc;
	
	FILE *plik=fopen("results","w");
	

	
	for (int n=1024;n<=32768;n+=1024) {
		experiment(n,1,pluginAlgs,3,plik,5);
	}
	
	for (int n=64;n<=1024;n+=64) {
		for (int d=1;d<=16;d++) {
			experiment(n,d,lscvAlgs,3,plik,5);
		}
	}
	
	for (int n=1024;n<=16384;n+=1024) {
		for (int d=1;d<=16;d++) {
			experiment(n,d,lscvMultiAlgs,3,plik,5);
		}
	}
	
	
	
	fclose(plik);
	
	return 0;
}
