#include "matrices.h"


float* transposeMatrix(float* m, unsigned int rows, unsigned int cols) {
	float *m2=new float[rows*cols];
	
	for (int i=0;i<rows;i++) {
		for (int j=0;j<cols;j++) {
			m2[i+j*rows]=m[j+i*cols];
		}
	}
	
	return m2;
}


float* covarianceMatrix(float* m, uint rows, uint cols) {
	float* o=new float[rows*rows];
	for (int i=0;i<rows;i++) {
		for (int j=0;j<=i-1;j++) {
			float sum1=0;
			float sum2=0;
			float sum3=0;
			for (int k=0;k<cols;k++) {
				sum1+=m[k+i*cols]*m[k+j*cols];
				sum2+=m[k+i*cols];
				sum3+=m[k+j*cols];
			}
			o[j+i*rows]=sum1/((float)cols-1.0f)-sum2*sum3/((float)cols*((float)cols-1.0f));
			o[i+j*rows]=o[j+i*rows];
		}
	}

	for (int i=0;i<rows;i++) {
		float sum1=0;
		float sum2=0;
		for (int k=0;k<cols;k++) {
			sum1+=m[k+i*cols]*m[k+i*cols];
			sum2+=m[k+i*cols];
		}

		o[i+i*rows]=sum1/((float)cols-1.0f)-sum2*sum2/((float)cols*((float)cols-1.0f));
	}

	return o;
}



void lupMatrixDecomposition(float *m, uint n, float **_o, int **_p, int* c) {
	float* o=new float[n*n];
	int* pi=new int[n];

	memcpy(o,m,sizeof(float)*n*n);
	for (int i=0;i<n;i++) pi[i]=i;

	*c=0;
	for (int k=0;k<n;k++) {
		float p=0;
		int kp=-1;
		for (int i=k;i<n;i++) {
			if (fabs(o[k+i*n])>p) {
				p=fabs(o[k+i*n]);
				kp=i;
			}
		}
		if (p==0) {
			printf("Macierz osobliwa \n");
			delete []o;
			delete []pi;
			break;
		}
		int tmp=pi[k];
		pi[k]=pi[kp];
		pi[kp]=tmp;

		*c+=(kp!=k);

		for (int i=0;i<n;i++) {
			float tmp1=o[i+k*n];
			o[i+k*n]=o[i+kp*n];
			o[i+kp*n]=tmp1;
		}
		for (int i=k+1;i<n;i++) {
			o[k+i*n]/=o[k+k*n];
			for (int j=k+1;j<n;j++) {
				o[j+i*n]=o[j+i*n]-o[k+i*n]*o[j+k*n];
			}
		}
	}

	*_o=o;
	*_p=pi;
}



void invertLowerTriangularMatrixWithoutAlloc(uint n, float *m, float *o) {
	for (int r=0;r<n;r++) {
		o[r+r*n]=1;
	}
	for (int r=1;r<n;r++) {
		o[r-1+r*n]=-m[r-1+r*n];
	}
	for (int r=2;r<n;r++) {
		o[r-2+r*n]=m[r-2+(r-1)*n]*m[r-2+1+r*n]-m[r-2+r*n];
	}
	for (int k=3;k<n;k++) {
		for (int r=k;r<n;r++) {
			float sum=0;
			//printf(" %d %d : \n",r,r-k);
			for(int i=0;i<k;i++) {
				sum+=m[r-k+(r-k+1+i)*n]*o[r-k+1+i+r*n];
				//printf("   (%d,%d)*(%d,%d) \n",r-k+1+i,r-k,r,r-k+1+i);
			}
			o[r-k+r*n]=-sum;
		}
	}
}

void invertUpperTriangularMatrixWithoutAlloc(uint n, float *m, float *o) {//Powstało poprzez transpozycję wszystkich współrzędnych w funkcji invertLowerTriangularMatrix

	for (int r=0;r<n;r++) {
		o[r+r*n]=1;
	}
	for (int r=1;r<n;r++) {
		o[r+(r-1)*n]=-m[r+(r-1)*n];
	}
	for (int r=2;r<n;r++) {
		o[r+(r-2)*n]=m[r-1+(r-2)*n]*m[r+(r-2+1)*n]-m[r+(r-2)*n];
	}
	for (int k=3;k<n;k++) {
		for (int r=k;r<n;r++) {
			float sum=0;
			//printf(" %d %d : \n",r,r-k);
			for(int i=0;i<k;i++) {
				sum+=m[r-k+1+i+(r-k)*n]*o[r+(r-k+1+i)*n];
				//printf("   (%d,%d)*(%d,%d) \n",r-k+1+i,r-k,r,r-k+1+i);
			}
			o[r+(r-k)*n]=-sum;
		}
	}

}



void invertLUPResultAndFindDeterminant(uint n, float *lu, int *p, float** inv, float *det, int c) {

	float *d=new float[n];
	float *_lu=new float[n*n];
	float *t=new float[n*n];
	float *o=new float[n*n];


	*det=1;

	for (int i=0;i<n;i++) {
		*det*=lu[i+i*n];
		d[i]=lu[i+i*n];
		for (int j=i;j<n;j++) {
			_lu[j+i*n]=lu[j+i*n]/d[i];
		}
		for (int j=0;j<i;j++) {
			_lu[j+i*n]=lu[j+i*n];
		}

	}

	*det*=(1-2*(c%2));

	invertUpperTriangularMatrixWithoutAlloc(n,_lu,t);

	invertLowerTriangularMatrixWithoutAlloc(n,_lu,t);

	for (int r=0;r<n;r++) {
		for (int c=0;c<n;c++) {
			o[c+r*n]=0;
			for (int i=myMax(r,c);i<n;i++) {
				o[c+r*n]+=t[i+r*n]*t[c+i*n]/d[i];
			}
		}
	}

	for (int r=0;r<n;r++) {
		for (int c=0;c<n;c++) {
			_lu[p[c]+r*n]=o[c+r*n];
		}
	}

	*inv=_lu;

	delete []d;
	delete []t;
	delete []o;
}


void invertMatrixAndFindDeterminant(uint n, float *m, float** _o, float *_det) {
	int *p;
	int c;
	float *inv;
	float *lu;
	float determinant;

	lupMatrixDecomposition(m, n, &lu, &p,&c);
	invertLUPResultAndFindDeterminant(n, lu, p, &inv, &determinant,c);

	delete []p;
	delete []lu;

	*_det=determinant;
	*_o=inv;
	/*alglib::real_2d_array a;
	a.setlength(n,n);
	
	for (int i=0;i<a.rows();i++) {
	  for (int j=0;j<a.cols();j++) {
	    a(i,j)=m[j+i*a.cols()];
	  }
	} 
	
	*_det=alglib::rmatrixdet(a);
	
	alglib::ae_int_t info;
	alglib::matinvreport rep;
	
	alglib::rmatrixinverse(a,info,rep);
	
	float* res=new float[n*n];
	
	for (int i=0;i<a.rows();i++) {
	  for (int j=0;j<a.cols();j++) {
	    res[j+i*a.cols()]=(float)a(i,j);
	  }
	} 
	
	
	*_o=res;*/
}

float* sqrtMatrix(float *m, uint n) {
	//Convert matrix m to the alglib format
	//Loop with assignment is used to achieve type conversion (in case when float!=double).
	alglib::real_2d_array a;
	a.setlength(n,n);
	
	for (int i=0;i<a.rows();i++) {
	  for (int j=0;j<a.cols();j++) {
	    a(i,j)=m[j+i*a.cols()];
	  }
	} 
	
	alglib::real_1d_array d;
	alglib::real_2d_array z;
	
	alglib::smatrixevd(a,n,1,true,d,z);
	
	float *mz=new float[n*n];
	
	//printf("mz \n");
	for (int i=0;i<z.rows();i++) {
	  for (int j=0;j<z.cols();j++) {
	    mz[j+i*z.cols()]=z(i,j);
	   // printf("%f ",z(i,j));
	  }
	  //printf("\n");
	} 
	
	float *invmz;
	float det;
	
	invertMatrixAndFindDeterminant(n, mz, &invmz, &det);
	
	/*printf("invmz \n");
	for (int i=0;i<n;i++) {
	  for (int j=0;j<n;j++) {
	    printf("%f ",invmz[j+i*n]);
	  }
	  printf("\n");
	} */
	
	float* result=new float[n*n];
	
	for (int i=0;i<n;i++) d(i)=sqrt(d(i));
	
	for (int i=0;i<n;i++) {//rows
		for(int j=0;j<n;j++) {//cols
			result[j+i*n]=0.0;
			for (int k=0;k<n;k++) {
				result[j+i*n]+=mz[k+i*n]*d(k)*invmz[j+k*n];
			}
		}
	}
		
	delete []mz;
	delete []invmz;
	
	return result;
}


void multMatrix(float *m,int cols, int rows,float coeff) {
	for (int i=0;i<rows;i++) {
		for (int j=0;j<cols;j++) {
			m[j+i*cols]*=coeff;
		}
	}
}

void multMatrix(float *out,float *m1,float *m2, int n) {
	for (int i=0;i<n;i++) {
		for (int j=0;j<n;j++) {
			out[j+i*n]=0.0;
			for (int k=0;k<n;k++) {
				out[j+i*n]+=m1[k+i*n]*m2[j+k*n];
			}
		}
	}
}
