#include "pluginsimple.h"


float PluginSimple::spk6c(float &x,int i,void *v) {
	float g1=((float*)v)[0];
	float xi=((float*)v)[1];
	
	return k6c((xi-x)/g1);
}

float PluginSimple::spk4c(float &x,int i,void *v) {
	float g2=((float*)v)[0];
	float xi=((float*)v)[1];
	
	return k4c((xi-x)/g2);
}

float PluginSimple::sqr(float &x,int i,void *v) {
	return x*x;
}


float PluginSimple::findH() {
	TIMESTAMP totalParallelizable=0;
	TIMESTAMP tempStart;
	TIMESTAMP start=getTimer();
	int n=getN();
	float *X=getData();
	
	 
	tempStart=getTimer();
	float sumx=stableUnoptimizedSum<true>(X,n);
	//printf("CPU sumX %f \n",sumx);
	float sumx2=stableUnoptimizedSumFun<PluginSimple::sqr,true>(X,n,NULL);
	totalParallelizable+=getTimer()-tempStart;
	//printf("CPU sumX2 %f \n",sumx2);
	
	float var=estVar(n,sumx,sumx2);
	float stdDev=estStdDev(var);
	float _psi8=psi8(stdDev);
	float g1=f4(_psi8,n);
	
	float *sumtemp=new float[n];
	
	float data[2];
	data[0]=g1;
	
	tempStart=getTimer();
	for (int i=1;i<n;i++) {
		data[1]=X[i];
		sumtemp[i-1]=stableUnoptimizedSumFun<PluginSimple::spk6c,true>(X,i,data);
	}
	
	
	float sumk6=stableUnoptimizedSum<false>(sumtemp,n-1);
	
	totalParallelizable+=getTimer()-tempStart;
	
	float _psi6=psi6(g1,n,sumk6);
	
	float g2=f2(_psi6,n);
	
	data[0]=g2;
	
	tempStart=getTimer();
	for (int i=1;i<n;i++) {
		data[1]=X[i];
		sumtemp[i-1]=stableUnoptimizedSumFun<PluginSimple::spk4c,true>(X,i,data);
	}
	
	float sumk4=stableUnoptimizedSum<false>(sumtemp,n-1);
	
	totalParallelizable+=getTimer()-tempStart;
	
	delete []sumtemp;
	
	float _psi4=psi4(g2,n,sumk4);
	float h=hcoeff(_psi4,n);
	
	TIMESTAMP end=getTimer();
	
	addMeasure("PLUGIN_SIMPLE_TOTAL",end-start);
	addMeasure("PLUGIN_SIMPLE_PARALLELIZABLE",totalParallelizable);
	
	return h;
}


void PluginSimple::startExp() {
	findH();
}
