#ifndef SIMPLEMATH_H
#define SIMPLEMATH_H

#include "constants.h"
#include "matrices.h"

unsigned int intDivCeil(unsigned int value,unsigned int divisor);
void vecvecMatBigSimple(float *X, float *m, int n, int d, float *out);
unsigned int log2i(unsigned int v);
unsigned int roundp2(unsigned int v);
float* findExponents(unsigned int d, unsigned int n,float* data,float* invCov);

inline float kstar(float h2,unsigned int n,float C1, float C2,float* exponents,unsigned int i) {
	float k;
	float k2;

	k=exp(-exponents[i]/(2.0f*h2))/C1;
	k2=exp(-exponents[i]/(4.0f*h2))/C2;

	return k2-2*k;
}



inline int mysqrt(int a) {
	float f=sqrt(a);
	int p=round(f);
	return (p*p==a)?p:floor(f);
}

inline float k6c(float x) {
	float x2=x*x;
	return k_coeff*exp(-0.5f*x2)*(x2*(x2*(x2-15.0f)+45.0f)-15.0f);
}

inline float k4c(float x) {
	float x2=x*x;
	return k_coeff*exp(-0.5f*x2)*(x2*(x2-6.0f)+3.0f);
}

inline float estVar(float n, float sumx, float sumx2) {
	float t1=1.0/(n-1.0);
	float t2=1.0/(n*(n-1));
	return t1*sumx2-t2*sumx*sumx;
}

inline float estStdDev(float var) {
	return sqrt(var);
}

inline float psi8(float stdDev) {
	return 105.0/(32.0*sqrt(PI)*pow(stdDev,9.0));
}

inline float f4(float _psi8,float n) {
	return pow(-2*k6_0/(mi2k*_psi8*n),1.0/9.0);
}


inline float psi6(float g1,float n, float sumk6) {
	return (2.0f*sumk6+n*k6_0)/(n*n*pow(g1,7.0f));
}

inline float f2(float _psi6, float n) {
	return pow(-2*k4_0/(mi2k*_psi6*n),1.0/7.0);
}


inline float psi4(float g2,float n, float sumk4) {
	return (2.0f*sumk4+n*k4_0)/(n*n*pow(g2,5.0));
}

inline float hcoeff(float _psi4,float n) {
	return pow(rk/(mi2k*mi2k*_psi4*n),1.0/5.0);
}

template<bool preserve> float stableUnoptimizedSum(float *t,int n) {
	float *temp;
	unsigned int n2=roundp2(n); //zaokrąglij do pierwszej większej potęgi 2.
	
	if (n2==1) return t[0];
	
	if (preserve) {
		temp=new float[n2/2];
	} else {
		temp=t;
	}
	
	for (int i=0;i<n2/2;i++) {
		if ((i+n2/2)<n) {
			temp[i]=t[i]+t[i+n2/2];
		} else {
			temp[i]=t[i];
		}
	}
	for (int step=n2/4;step>=1;step=step/2) {
		for (int i=0;i<step;i++) {
			if ((i+step)<n) temp[i]=temp[i]+temp[i+step];
		}
	}
	
	float result=temp[0];
	
	if (preserve) {
		delete []temp;
	}
	
	return result;
}


template <float (*F)(float &,int i,void *),bool preserve> float stableUnoptimizedSumFun(float *t,int n,void *data) {
	
	float *temp;
	unsigned int n2=roundp2(n); //zaokrąglij do pierwszej większej potęgi 2.
	
	if (n2==1) return F(t[0],0,data);
	
	if (preserve) {
		temp=new float[n2/2];
	} else {
		temp=t;
	}

	for (int i=0;i<n2/2;i++) {
		if ((i+n2/2)<n) {
			temp[i]=F(t[i],i,data)+F(t[i+n2/2],i+n2/2,data);
		} else {
			temp[i]=F(t[i],i,data);
		}
	}
	for (int step=n2/4;step>=1;step=step/2) {
		for (int i=0;i<step;i++) {
			if ((i+step)<n) temp[i]=temp[i]+temp[i+step];
		}
	}
	
	float result=temp[0];
	
	if (preserve) {
		delete []temp;
	}
	
	return result;
}


#endif