#include <ssemath.h>


void vecvecMat16(float *X,float *m,int q,int l,int n,int d,__m128 *out);
void vecvecMatTri16(float *X,float *m,int s,int n,int d,__m128 *out);


//Interface functions
__m128 sseDotProduct(__m128 &l, __m128 *p) {
	__m128 r1=_mm_dp_ps(l,p[0],0xf1);
	__m128 r2=_mm_dp_ps(l,p[1],0xf2);
	r1=_mm_add_ps(r1,r2);
	__m128 r3=_mm_dp_ps(l,p[2],0xf4);
	__m128 r4=_mm_dp_ps(l,p[3],0xf8);
	r3=_mm_add_ps(r3,r4);
	return _mm_add_ps(r1,r3);
}

void transpose4x4(__m128 &r1,__m128 &r2,__m128 &r3,__m128 &r4) {
	__m128 v1,v2,v3,v4;
	
	v1=_mm_unpacklo_ps(r1,r3);
	v2=_mm_unpackhi_ps(r1,r3);
	v3=_mm_unpacklo_ps(r2,r4);
	v4=_mm_unpackhi_ps(r2,r4);
	
	r1=_mm_unpacklo_ps(v1,v3);
	r2=_mm_unpackhi_ps(v1,v3);
	r3=_mm_unpacklo_ps(v2,v4);
	r4=_mm_unpackhi_ps(v2,v4);
}

void vecvecMatBigSSEParallel(float *X,float *m,int n,int d,float *out,int threads) {
	int i;
	int s=n>>4;//n/16
	int blocks=(s*(s-1))>>1;
	omp_set_num_threads(threads);
	
	#pragma omp parallel shared(X,m,n,d) private(i)
	{
		#pragma omp for schedule(auto) nowait
		for (i=0;i<blocks;i++) {
			int l=(1+mysqrt(1+i<<3))>>1;
			int q=i-l*(l-1)/2;
			l=l<<4;
			q=q<<4;
			vecvecMat16(X,m,q,l,n,d,(__m128*)(out+256*i));
		}
		#pragma omp for schedule(auto) nowait
		for (i=0;i<s;i++) {
			vecvecMatTri16(X,m,i<<4,n,d,(__m128*)(out+256*blocks+i*120));
		}
	}
	
	
}



//Auxiliary functions

float aggregatePow2(float *t,float* temp,int n) { //n has to be a power of 2 and greater than 8
	unsigned int iter=log2i(n)-2;
	
	if ((n&(n-1))!=0) {
		printf("n nie jest potega 2!");
	}
	
	if (n<8) {
		printf("n<8 \n");
	}
	
	n=n>>3;// /2 bo zwijamy 2 w jeden, /4 bo zwijamy w paczkach po 4
	
	for (unsigned int i=0;i<n;i++) {
		__m128 *a=(((__m128*)t)+(i<<1));
		__m128 *b=(((__m128*)t)+((i<<1)+1));
			
		*(((__m128*)temp)+i)=_mm_hadd_ps(*a,*b);
	}
	n=n>>1;
	for (unsigned int j=1;j<iter;j++) {
		for (unsigned int i=0;i<n;i++) {
			__m128 *a=(((__m128*)temp)+(i<<1));
			__m128 *b=(((__m128*)temp)+((i<<1)+1));
			
			*(((__m128*)temp)+i)=_mm_hadd_ps(*a,*b);
		}
		n=n>>1;
	}
	
	return temp[0]+temp[1]+temp[2]+temp[3];
}


void vecvecMat16(float *X,float *m,int q,int l,int n,int d,__m128 *out) {
	
	__m128 mq[4*16];
	__m128 ml[4*16];
	
	//Skopiuj dane do buforów, hopefully w cache'u
	for (int i=0;i<d;i++) {
		mq[(i<<2)+0]=*(__m128*)(X+n*i+q+0);
		mq[(i<<2)+1]=*(__m128*)(X+n*i+q+4);
		mq[(i<<2)+2]=*(__m128*)(X+n*i+q+8);
		mq[(i<<2)+3]=*(__m128*)(X+n*i+q+12);
		
		ml[(i<<2)+0]=*(__m128*)(X+n*i+l+0);
		ml[(i<<2)+1]=*(__m128*)(X+n*i+l+4);
		ml[(i<<2)+2]=*(__m128*)(X+n*i+l+8);
		ml[(i<<2)+3]=*(__m128*)(X+n*i+l+12);
	}
	
	__m128 res1;
	__m128 res2;
	__m128 res3;
	__m128 res4;
	__m128 part1;
	__m128 part2;
	__m128 part3;
	__m128 part4;
	__m128 c;
	__m128 v4;
	float v;
	
	for (int i=0;i<16;i++) {
		res1=_mm_set1_ps(0.0f);
		res2=res1;
		res3=res1;
		res4=res1;
		for (int p=0;p<d;p++) {
			part1=_mm_set1_ps(0.0f);
			part2=part1;
			part3=part1;
			part4=part1;
			for (int k=0;k<d;k++) {
				c=_mm_set1_ps(m[k*d+p]);
				v=((float*)ml)[k*16+i];
				v4=_mm_set1_ps(v);
				
				part1=_mm_add_ps(part1,_mm_mul_ps(c,_mm_sub_ps(v4,mq[k*4+0]))); //k*4 bo *16 i /4
				part2=_mm_add_ps(part2,_mm_mul_ps(c,_mm_sub_ps(v4,mq[k*4+1]))); //k*4 bo *16 i /4
				part3=_mm_add_ps(part3,_mm_mul_ps(c,_mm_sub_ps(v4,mq[k*4+2]))); //k*4 bo *16 i /4
				part4=_mm_add_ps(part4,_mm_mul_ps(c,_mm_sub_ps(v4,mq[k*4+3]))); //k*4 bo *16 i /4
			}
		
			v=((float*)ml)[p*16+i];
			v4=_mm_set1_ps(v);
			res1=_mm_add_ps(res1,_mm_mul_ps(part1,_mm_sub_ps(v4,mq[p*4+0])));
			res2=_mm_add_ps(res2,_mm_mul_ps(part2,_mm_sub_ps(v4,mq[p*4+1])));
			res3=_mm_add_ps(res3,_mm_mul_ps(part3,_mm_sub_ps(v4,mq[p*4+2])));
			res4=_mm_add_ps(res4,_mm_mul_ps(part4,_mm_sub_ps(v4,mq[p*4+3])));
			//wyniki są generowane wierszami
		}	
		
		out[0]=res1;
		out[1]=res2;
		out[2]=res3;
		out[3]=res4;
		out+=4;
/*				float *b=(float*)&res1;
				printf("%.0f %.0f %.0f %.0f ",b[0],b[1],b[2],b[3]);
				b=(float*)&res2;
				printf("%.0f %.0f %.0f %.0f ",b[0],b[1],b[2],b[3]);
				b=(float*)&res3;
				printf("%.0f %.0f %.0f %.0f ",b[0],b[1],b[2],b[3]);
				b=(float*)&res4;
				printf("%.0f %.0f %.0f %.0f\n",b[0],b[1],b[2],b[3]);
			
*/		
			
	}
}

void vecvecMatTri16(float *X,float *m,int s,int n,int d,__m128 *out) {
	__m128 ms[4*16];
	
	//Skopiuj dane do buforów, hopefully w cache'u
	for (int i=0;i<d;i++) {
		//printf("n %d, i %d, s %d, addr %d %d %d %d \n",n,i,s,n*i+s,n*i+s+4,n*i+s+8,n*i+s+12);
		ms[(i<<2)+0]=*(__m128*)(X+n*i+s+0);
		ms[(i<<2)+1]=*(__m128*)(X+n*i+s+4);
		ms[(i<<2)+2]=*(__m128*)(X+n*i+s+8);
		ms[(i<<2)+3]=*(__m128*)(X+n*i+s+12);
	}
	
	__m128 res1;
	__m128 res2;
	__m128 res3;
	__m128 res4;
	__m128 part1;
	__m128 part2;
	__m128 part3;
	__m128 part4;
	__m128 c;
	__m128 v4;
	__m128 temp[4];
	float v;
	
	float* myout=(float*) out;
	for (int i=1;i<16;i++) {
		res1=_mm_set1_ps(0.0f);
		res2=res1;
		res3=res1;
		res4=res1;
		for (int p=0;p<d;p++) {
			part1=_mm_set1_ps(0.0f);
			part2=part1;
			part3=part1;
			part4=part1;
			for (int k=0;k<d;k++) {
				c=_mm_set1_ps(m[k*d+p]);
				v=((float*)ms)[k*16+i];
				v4=_mm_set1_ps(v);
				
				part1=_mm_add_ps(part1,_mm_mul_ps(c,_mm_sub_ps(v4,ms[k*4+0]))); //k*4 bo *16 i /4
				part2=_mm_add_ps(part2,_mm_mul_ps(c,_mm_sub_ps(v4,ms[k*4+1]))); //k*4 bo *16 i /4
				part3=_mm_add_ps(part3,_mm_mul_ps(c,_mm_sub_ps(v4,ms[k*4+2]))); //k*4 bo *16 i /4
				part4=_mm_add_ps(part4,_mm_mul_ps(c,_mm_sub_ps(v4,ms[k*4+3]))); //k*4 bo *16 i /4
			}
		
			v=((float*)ms)[p*16+i];
			v4=_mm_set1_ps(v);
			res1=_mm_add_ps(res1,_mm_mul_ps(part1,_mm_sub_ps(v4,ms[p*4+0])));
			res2=_mm_add_ps(res2,_mm_mul_ps(part2,_mm_sub_ps(v4,ms[p*4+1])));
			res3=_mm_add_ps(res3,_mm_mul_ps(part3,_mm_sub_ps(v4,ms[p*4+2])));
			res4=_mm_add_ps(res4,_mm_mul_ps(part4,_mm_sub_ps(v4,ms[p*4+3])));
			//wyniki są generowane wierszami
		}	
		
		temp[0]=res1;
		temp[1]=res2;
		temp[2]=res3;
		temp[3]=res4;
		
		for (int p=0;p<i;p++) {
			myout[p]=((float*)temp)[p];
			//printf("%d:%.0f ",p,((float*)temp)[p]);
		}
		myout=myout+i;
		//printf("\n");
					
	}
	
}


