#ifndef SSEMATH_H
#define SSEMATH_H

#include <math.h>
#include <omp.h>
#include "smmintrin.h"
#include "simplemath.h"
#include <stdio.h>
#include "utils.h"

__m128 sseDotProduct(__m128 &l, __m128 *p);
void transpose4x4(__m128 &r1,__m128 &r2,__m128 &r3,__m128 &r4);
void vecvecMatBigSSEParallel(float *X,float *m,int n,int d,float *out,int threads=16);
float aggregatePow2(float *t,int n);
float multiThreadedAggregateOMP(float *t, int n, int threads=4,int cacheSize=1<<15);
float aggregatePow2(float *t,float* temp,int n);




template <__m128 (*F1)(__m128,void*)> float vecvecMat16FunSum(float *X,float *m,int q,int l,int n,int d,void* data) {
	
	__m128 mq[4*16];
	__m128 ml[4*16];
	
	//Skopiuj dane do buforów, hopefully w cache'u
	for (int i=0;i<d;i++) {
		mq[(i<<2)+0]=*(__m128*)(X+n*i+q+0);
		mq[(i<<2)+1]=*(__m128*)(X+n*i+q+4);
		mq[(i<<2)+2]=*(__m128*)(X+n*i+q+8);
		mq[(i<<2)+3]=*(__m128*)(X+n*i+q+12);
		
		ml[(i<<2)+0]=*(__m128*)(X+n*i+l+0);
		ml[(i<<2)+1]=*(__m128*)(X+n*i+l+4);
		ml[(i<<2)+2]=*(__m128*)(X+n*i+l+8);
		ml[(i<<2)+3]=*(__m128*)(X+n*i+l+12);
	}
	
	__m128 res1;
	__m128 res2;
	__m128 res3;
	__m128 res4;
	__m128 part1;
	__m128 part2;
	__m128 part3;
	__m128 part4;
	__m128 c;
	__m128 v4;
	__m128 res=_mm_set1_ps(0.0f);
	float v;
	
	for (int i=0;i<16;i++) {
		res1=_mm_set1_ps(0.0f);
		res2=res1;
		res3=res1;
		res4=res1;
		for (int p=0;p<d;p++) {
			part1=_mm_set1_ps(0.0f);
			part2=part1;
			part3=part1;
			part4=part1;
			for (int k=0;k<d;k++) {
				c=_mm_set1_ps(m[k*d+p]);
				v=((float*)ml)[k*16+i];
				v4=_mm_set1_ps(v);
				
				part1=_mm_add_ps(part1,_mm_mul_ps(c,_mm_sub_ps(v4,mq[k*4+0]))); //k*4 bo *16 i /4
				part2=_mm_add_ps(part2,_mm_mul_ps(c,_mm_sub_ps(v4,mq[k*4+1]))); //k*4 bo *16 i /4
				part3=_mm_add_ps(part3,_mm_mul_ps(c,_mm_sub_ps(v4,mq[k*4+2]))); //k*4 bo *16 i /4
				part4=_mm_add_ps(part4,_mm_mul_ps(c,_mm_sub_ps(v4,mq[k*4+3]))); //k*4 bo *16 i /4
			}
		
			v=((float*)ml)[p*16+i];
			v4=_mm_set1_ps(v);
			res1=_mm_add_ps(res1,_mm_mul_ps(part1,_mm_sub_ps(v4,mq[p*4+0])));
			res2=_mm_add_ps(res2,_mm_mul_ps(part2,_mm_sub_ps(v4,mq[p*4+1])));
			res3=_mm_add_ps(res3,_mm_mul_ps(part3,_mm_sub_ps(v4,mq[p*4+2])));
			res4=_mm_add_ps(res4,_mm_mul_ps(part4,_mm_sub_ps(v4,mq[p*4+3])));
			//wyniki są generowane wierszami
		}	
		res1=F1(res1,data);
		res2=F1(res2,data);
		res3=F1(res3,data);
		res4=F1(res4,data);
		res=_mm_add_ps(res,_mm_add_ps(_mm_add_ps(res1,res2),_mm_add_ps(res3,res4)));
	}
	
	float *t=(float*)&res;
	
	return t[0]+t[1]+t[2]+t[3];
}

template <__m128 (*F1)(__m128,void*)> float vecvecMatTri16FunSum(float *X,float *m,int s,int n,int d,void* data) {
	__m128 ms[4*16];
	
	//Skopiuj dane do buforów, hopefully w cache'u
	for (int i=0;i<d;i++) {
		//printf("n %d, i %d, s %d, addr %d %d %d %d \n",n,i,s,n*i+s,n*i+s+4,n*i+s+8,n*i+s+12);
		ms[(i<<2)+0]=*(__m128*)(X+n*i+s+0);
		ms[(i<<2)+1]=*(__m128*)(X+n*i+s+4);
		ms[(i<<2)+2]=*(__m128*)(X+n*i+s+8);
		ms[(i<<2)+3]=*(__m128*)(X+n*i+s+12);
	}
	
	__m128 res1;
	__m128 res2;
	__m128 res3;
	__m128 res4;
	__m128 part1;
	__m128 part2;
	__m128 part3;
	__m128 part4;
	__m128 last;
	__m128 res=_mm_set1_ps(0.0f);
	__m128 c;
	__m128 v4;
	//__m128 temp[4];
	float v;
	
	//float* myout=(float*) out;
	for (int i=1;i<16;i++) {
		res1=_mm_set1_ps(0.0f);
		res2=res1;
		res3=res1;
		res4=res1;
		for (int p=0;p<d;p++) {
			part1=_mm_set1_ps(0.0f);
			part2=part1;
			part3=part1;
			part4=part1;
			for (int k=0;k<d;k++) {
				c=_mm_set1_ps(m[k*d+p]);
				v=((float*)ms)[k*16+i];
				v4=_mm_set1_ps(v);
				
				part1=_mm_add_ps(part1,_mm_mul_ps(c,_mm_sub_ps(v4,ms[k*4+0]))); //k*4 bo *16 i /4
				part2=_mm_add_ps(part2,_mm_mul_ps(c,_mm_sub_ps(v4,ms[k*4+1]))); //k*4 bo *16 i /4
				part3=_mm_add_ps(part3,_mm_mul_ps(c,_mm_sub_ps(v4,ms[k*4+2]))); //k*4 bo *16 i /4
				part4=_mm_add_ps(part4,_mm_mul_ps(c,_mm_sub_ps(v4,ms[k*4+3]))); //k*4 bo *16 i /4
			}
		
			v=((float*)ms)[p*16+i];
			v4=_mm_set1_ps(v);
			res1=_mm_add_ps(res1,_mm_mul_ps(part1,_mm_sub_ps(v4,ms[p*4+0])));
			res2=_mm_add_ps(res2,_mm_mul_ps(part2,_mm_sub_ps(v4,ms[p*4+1])));
			res3=_mm_add_ps(res3,_mm_mul_ps(part3,_mm_sub_ps(v4,ms[p*4+2])));
			res4=_mm_add_ps(res4,_mm_mul_ps(part4,_mm_sub_ps(v4,ms[p*4+3])));
			//wyniki są generowane wierszami
		}	
		
		
		
		
		res1=F1(res1,data);
		res2=F1(res2,data);
		res3=F1(res3,data);
		res4=F1(res4,data);
		
		//__debug code start
		
		/*__m128 temp[4];
		temp[0]=res1;
		temp[1]=res2;
		temp[2]=res3;
		temp[3]=res4;
		for (int p=0;p<i;p++) printf(">> %f \n",((float*)temp)[p]);*/
		//__debug code end
		
		
		last=res1;
		if (i>4) {
			res=_mm_add_ps(res,res1);
			last=res2;
		}
		if (i>8) {
			res=_mm_add_ps(res,res2);
			last=res3;
		}
		if (i>12) {
			res=_mm_add_ps(res,res3);
			last=res4;
		}
		
		if ((i&3)==1) last=_mm_mul_ps(last,_mm_set_ps(0.0f,0.0f,0.0f,1.0f));
		if ((i&3)==2) last=_mm_mul_ps(last,_mm_set_ps(0.0f,0.0f,1.0f,1.0f));
		if ((i&3)==3) last=_mm_mul_ps(last,_mm_set_ps(0.0f,1.0f,1.0f,1.0f));
		
		res=_mm_add_ps(res,last);
					
	}
	
	float *t=(float*)&res;
	
	return t[0]+t[1]+t[2]+t[3];
	
}

template <__m128 (*F1)(__m128,void*)> float vecvecMatBigSSEParallelFunSum(float *X,float *m,int n,int d,int threads,void* data) {
	int i;
	int s=n>>4;//n/16
	int blocks=(s*(s-1))>>1;
	float result=0.0f;
	omp_set_num_threads(threads);
	
	#pragma omp parallel shared(X,m,n,d) private(i) reduction(+ : result)
	{
		#pragma omp for schedule(auto) nowait
		for (i=0;i<blocks;i++) {
			int l=(1+mysqrt(1+i<<3))>>1;
			int q=i-l*(l-1)/2;
			l=l<<4;
			q=q<<4;
			result+=vecvecMat16FunSum<F1>(X,m,q,l,n,d,data);
		}
		#pragma omp for schedule(auto) nowait
		for (i=0;i<s;i++) {
			result+=vecvecMatTri16FunSum<F1>(X,m,i<<4,n,d,data);
		}
	}
	
	return result;
}

template <__m128 (*F1)(__m128)> float vecvecMat1dFun16(float *X,int q,int l,float g) {
	__m128 mq[16];
	__m128 ml[16];
	
	//Skopiuj dane do buforów, hopefully w cache'u
	for (int i=0;i<4;i++) mq[i]=*(__m128*)(X+q+(i<<2)); 
	for (int i=0;i<4;i++) ml[i]=*(__m128*)(X+l+(i<<2)); 
	
	__m128 total=_mm_set_ps(0.0f,0.0f,0.0f,0.0f);
	__m128 gv=_mm_set_ps(g,g,g,g);
	
	for (int i=0;i<4;i++) {
		__m128 val=mq[i];
		
		__m128 x1=_mm_shuffle_ps(val,val,_MM_SHUFFLE(0,0,0,0));
		__m128 x2=_mm_shuffle_ps(val,val,_MM_SHUFFLE(1,1,1,1));
		__m128 x3=_mm_shuffle_ps(val,val,_MM_SHUFFLE(2,2,2,2));
		__m128 x4=_mm_shuffle_ps(val,val,_MM_SHUFFLE(3,3,3,3));
		
		__m128 acc=_mm_set_ps(0.0f,0.0f,0.0f,0.0f);
		
		for (int j=0;j<4;j++) {
			acc=_mm_add_ps(acc,F1(_mm_div_ps(_mm_sub_ps(x1,ml[j]),gv)));
			acc=_mm_add_ps(acc,F1(_mm_div_ps(_mm_sub_ps(x2,ml[j]),gv)));
			acc=_mm_add_ps(acc,F1(_mm_div_ps(_mm_sub_ps(x3,ml[j]),gv)));
			acc=_mm_add_ps(acc,F1(_mm_div_ps(_mm_sub_ps(x4,ml[j]),gv)));
		}
		
		total=_mm_add_ps(total,acc);
	}
	
	float *t=(float*)&total;
	
	return t[0]+t[1]+t[2]+t[3];
}

template <__m128 (*F1)(__m128)> float vecvecMat1dFunTri16(float *X,int s,float g) {
	__m128 ms[16];
	
	//Skopiuj dane do buforów, hopefully w cache'u
	for (int i=0;i<4;i++) ms[i]=*(__m128*)(X+s+(i<<2)); 
	__m128 total=_mm_set_ps(0.0f,0.0f,0.0f,0.0f);
	__m128 gv=_mm_set_ps(g,g,g,g);
	
	
	for (int i=1;i<4;i++) {
		__m128 val=ms[i];
		
		__m128 x1=_mm_shuffle_ps(val,val,_MM_SHUFFLE(0,0,0,0));
		__m128 x2=_mm_shuffle_ps(val,val,_MM_SHUFFLE(1,1,1,1));
		__m128 x3=_mm_shuffle_ps(val,val,_MM_SHUFFLE(2,2,2,2));
		__m128 x4=_mm_shuffle_ps(val,val,_MM_SHUFFLE(3,3,3,3));
		
		__m128 acc=_mm_set_ps(0.0f,0.0f,0.0f,0.0f);
		
		for (int j=0;j<i;j++) {
			acc=_mm_add_ps(acc,F1(_mm_div_ps(_mm_sub_ps(x1,ms[j]),gv)));
			acc=_mm_add_ps(acc,F1(_mm_div_ps(_mm_sub_ps(x2,ms[j]),gv)));
			acc=_mm_add_ps(acc,F1(_mm_div_ps(_mm_sub_ps(x3,ms[j]),gv)));
			acc=_mm_add_ps(acc,F1(_mm_div_ps(_mm_sub_ps(x4,ms[j]),gv)));
		}
		
		total=_mm_add_ps(total,acc);
	}
	
	for (int i=0;i<4;i++) {
		__m128 val=ms[i];
		
		__m128 x1=_mm_shuffle_ps(val,val,_MM_SHUFFLE(0,0,0,0));
		__m128 x2=_mm_shuffle_ps(val,val,_MM_SHUFFLE(1,1,1,1));
		__m128 x3=_mm_shuffle_ps(val,val,_MM_SHUFFLE(2,2,2,2));
		
		__m128 acc=_mm_mul_ps(F1(_mm_div_ps(_mm_sub_ps(x1,ms[i]),gv)),_mm_set_ps(1.0,1.0,1.0,0.0));
		acc=_mm_add_ps(acc,_mm_mul_ps(F1(_mm_div_ps(_mm_sub_ps(x2,ms[i]),gv)),_mm_set_ps(1.0,1.0,0.0,0.0)));
		acc=_mm_add_ps(acc,_mm_mul_ps(F1(_mm_div_ps(_mm_sub_ps(x3,ms[i]),gv)),_mm_set_ps(1.0,0.0,0.0,0.0)));

		total=_mm_add_ps(total,acc);
	}
	
	float *t=(float*)&total;
	
	return t[0]+t[1]+t[2]+t[3];
	
}

template <__m128 (*F1)(__m128)> float vecvecMatBigSSE1dFun(float *X,int n,float g,int threads=8) {
	int i;
	int s=n>>4;//n/16
	int blocks=(s*(s-1))>>1;
	omp_set_num_threads(threads);
	
	float result=0.0f;
	
	#pragma omp parallel shared(X,n) private(i) reduction(+ : result)
	{
		#pragma omp for schedule(auto) nowait
		for (i=0;i<blocks;i++) {
			int l=(1+mysqrt(1+i<<3))>>1;
			int q=i-l*(l-1)/2;
			l=l<<4;
			q=q<<4;
			result+=vecvecMat1dFun16<F1>(X,q,l,g);
		}
		#pragma omp for schedule(auto) nowait
		for (i=0;i<s;i++) {
			result+=vecvecMat1dFunTri16<F1>(X,i<<4,g);
		}
	}
	
	return result;
	
}


template <bool preserve> float multiThreadedAggregateOMP(float *t, int n, int threads,int cacheSize) { 
	int nPerCache=cacheSize/sizeof(float);
	float result=0;
	int i;
	int chunks=n/nPerCache;
	omp_set_num_threads(threads);

	float *temp;
	
	if (preserve) {
		temp=new float[n];
	} else {
		temp=t;
	}
	
	#pragma omp parallel shared(t,n,chunks,nPerCache) private(i) reduction(+ : result)
	{
		#pragma omp for schedule(auto) nowait
		for (i=0;i<chunks;i++) {
			result+=aggregatePow2(t+i*nPerCache,temp+i*nPerCache,nPerCache);
		}
		
	}
	
	int left=(n-chunks*nPerCache)&(~(0x07));
	float *p=t+chunks*nPerCache;
	int maxbit=log2i(left);
	
	if (left>0) {
		omp_set_num_threads(myMin(threads,maxbit-3+1));
			
		#pragma omp parallel shared(t,n,maxbit,left,p) private(i) reduction(+ : result)
		{
			#pragma omp for schedule(auto) nowait
			for (i=3;i<=maxbit;i++) {
				if ((left & (1<<i))>0) {	
					result+=aggregatePow2(p+(left&((1<<i)-1)),temp+(left&((1<<i)-1)),1<<i);
				}
			}
		}
	}
	
	p+=left;
	left=n-chunks*nPerCache-left;
	
	for (int i=0;i<left;i++) result+=p[i];
	
	if (preserve) {
		delete []temp;
	} 
	
	return result;
}


//Aproksymacja szeregiem Taylora
template <int steps> inline __m128 _mm_exptaylor_ps(__m128 &x) {
	__m128 ones=_mm_set_ps(1.0f,1.0f,1.0f,1.0f);
	__m128 res=ones;
	__m128 divisor=ones;
	__m128 v=x;
	
	res=_mm_add_ps(res,v);
	
	for (int i=0;i<steps-1;i++) {
		divisor=_mm_add_ps(divisor,ones);
		v=_mm_div_ps(_mm_mul_ps(v,x),divisor);
		res=_mm_add_ps(res,v);
	}
	
	return res;
}

//Auxiliary function
template <__m128 (*F1)(__m128 &,int i,void *)> float aggregatePow2Fun(float *t,float *temp, int n,void* data) { //n has to be a power of 2 and greater or equal to 8
	unsigned int iter=log2i(n)-2;
	
	if ((n&(n-1))!=0) {
		printf("n nie jest potega 2!");
	}
	
	if (n<8) {
		printf("n<8 \n");
	}
	
	n=n>>3;// /2 bo zwijamy 2 w jeden, /4 bo zwijamy w paczkach po 4
	
	for (unsigned int i=0;i<n;i++) {
		__m128 a=*(((__m128*)t)+(i<<1));
		__m128 b=*(((__m128*)t)+((i<<1)+1));
			
		a= F1(a,i,data);
		b= F1(b,i,data);
			
		
		*(((__m128*)temp)+i)=_mm_hadd_ps(a,b);
	}
	n=n>>1;
	for (unsigned int j=1;j<iter;j++) {
		for (unsigned int i=0;i<n;i++) {
			__m128 *a=(((__m128*)temp)+(i<<1));
			__m128 *b=(((__m128*)temp)+((i<<1)+1));
			
			*(((__m128*)temp)+i)=_mm_hadd_ps(*a,*b);
		}
		n=n>>1;
	}
	
	return temp[0]+temp[1]+temp[2]+temp[3];
}

template <__m128 (*F1)(__m128 &,int i,void *),float (*F2)(float &,int i,void *),bool preserve> float multiThreadedAggregateOMPFun(float *t, int n, void* data,int threads=4,int cacheSize=1<<15) { 
	int nPerCache=cacheSize/sizeof(float);
	float result=0;
	int i;
	int chunks=n/nPerCache;
	omp_set_num_threads(threads);

	float *temp;
	
	if (preserve) {
		temp=new float[n];
	} else {
		temp=t;
	}
	
	#pragma omp parallel shared(t,n,chunks,nPerCache) private(i) reduction(+ : result)
	{
		#pragma omp for schedule(auto) nowait
		for (i=0;i<chunks;i++) {
			result+=aggregatePow2Fun<F1>(t+i*nPerCache,temp+i*nPerCache,nPerCache,data);
		}
		
		
	}
	
	int left=(n-chunks*nPerCache)&(~(0x07));
	float *p=t+chunks*nPerCache;
	int maxbit=log2i(left);
	
	if (left>0) {
		omp_set_num_threads(myMin(threads,maxbit-3+1));
			
		#pragma omp parallel shared(t,n,maxbit,left,p) private(i) reduction(+ : result)
		{
			#pragma omp for schedule(auto) nowait
			for (i=3;i<=maxbit;i++) {
				if ((left & (1<<i))>0) {	
					result+=aggregatePow2Fun<F1>(p+(left&((1<<i)-1)),temp+(left&((1<<i)-1)),1<<i,data);
				}
			}
		}
	}
	
	p+=left;
	left=n-chunks*nPerCache-left;
	for (int i=0;i<left;i++) result+=F2(p[i],i,data);
	
	if (preserve) {
		delete []temp;
	} 
	
	return result;
}




#endif